/** 
* @file TimeboyStation.h
* @brief Schnittstelle fuer die Klasse TimeboyStation.
*/
//////////////////////////////////////////////////////////////////////
#pragma once
#ifndef DATAFOX_COMMUNICATION_TIMEBOY_STATION_V1_H
#define DATAFOX_COMMUNICATION_TIMEBOY_STATION_V1_H

// Basisklasse einbinden.
#include "datafox/communication/connection_base.h"

#include <time.h>
#include <thread>

//######################## TIMEBOY SLOT #######################################
// Defines Protokollbefehle
#define MB_READ		0x21		// bis zu 64 Byte lesen
#define MB_WRITE	0x22		// bis zu 64 Byte schreiben

// Defines fuer bestimmte Daten zu einem Slot
#define MB_SLOT_CHECK                  0x8000000C // Slot parameter fuer De/aktivierung der Timeboyerkennung
#define MB_SLOT_LED                    0x80001000 // Rote LED ein/aus
// Schreibe 1 Byte mit Wert 0 fuer: ein
// Schreibe 1 Byte mit Wert 255 fuer: aus
#define MB_SLOT_STATE                  0x8000E000 // Aktueller Steckzustand im Slot
#define MB_SLOT_VERSION                0x8000FFFF // Softwareversion der Slotfirmware

#define MB_SLOT_BUFFER_SIZE_SMALL      45  // Paketpuffergroesse bei FWVersion < 1.31
#define MB_SLOT_BUFFER_SIZE_LARGE      150 // Paketpuffergroesse bei FWVersion >= 1.31
#define MB_SLOT_BUFFER_DATA_SIZE_SMALL	(MB_SLOT_BUFFER_SIZE_SMALL - 13)
#define MB_SLOT_BUFFER_DATA_SIZE_LARGE	(MB_SLOT_BUFFER_SIZE_LARGE - 13)
//#############################################################################

// Versenden und Empfangen von Datenpacketen Tracen
//#define TRACE_DOCKING 1

namespace Datafox
{
    namespace Communication
    {
        // Datenstrucktur mit Daten zu einem Slot.
        class Timeboy_slot  
        {
        public:
            Timeboy_slot()
            {
                reset();
            }

        public:
            void reset()
            {
                // Member initialisieren
                m_version = 0;
                m_detection_state = 0;
                m_led_state = 0;
                m_led_stamp = 0;
                m_receive_stamp_slot = 0;
                m_receive_stamp_timeboy = 0;
            }
        private:
            unsigned short m_version;          // Ermittelte Firmwareversion
        public:
            void set_version( unsigned short version )
            {
                m_version = version;
            }
            unsigned short get_version()
            {
                return m_version;
            }
        private:
            unsigned char m_detection_state;    // Ermittlung der Komperatorschaltung; 0 -> nicht ermittelt, 1 -> nicht verfuegbar, 2 -> verfuegbar
        public:
            void set_detection_state( unsigned char detection_state )
            {
                m_detection_state = detection_state;
            }
            unsigned char get_detection_state()
            {
                return m_detection_state;
            }
        private:
            unsigned char m_led_state;          // Zuletzt gesetzten LED-Zustand: 0x00 -> NA, 0x01 -> ON, 0xFF -> OFF
        public:
            void set_led_state( unsigned char led_state )
            {
                m_led_state = led_state;
            }
            unsigned char get_led_state()
            {
                return m_led_state;
            }
        private:
            clock_t m_led_stamp;                // clock() des letzten Zustand setzen.
        public:
            void set_led_stamp( clock_t led_stamp )
            {
                m_led_stamp = led_stamp;
            }
            clock_t get_led_stamp()
            {
                return m_led_stamp;
            }
        private:
            clock_t m_receive_stamp_slot;        // clock() des zuletzt gueltig empfangenen Pakets.
        public:
            void set_receive_stamp_slot( clock_t receive_stamp_slot )
            {
                m_receive_stamp_slot = receive_stamp_slot;
            }
            clock_t get_receive_stamp_slot()
            {
                return m_receive_stamp_slot;
            }
        private:
            clock_t m_receive_stamp_timeboy;     // clock() des zuletzt gueltig empfangene Antwort vom Timeboy
        public:
            void set_receive_stamp_timeboy( clock_t receive_stamp_timeboy )
            {
                m_receive_stamp_timeboy = receive_stamp_timeboy;
            }
            clock_t get_receive_stamp_timeboy()
            {
                return m_receive_stamp_timeboy;
            }

        };

// Automatische Jobs des Kommunikationsworkers
#define TSF_AUTO_LED         fb_1 // Ausschalten der Slot-LED durch Worker, nach vorgegebenem Timeout
#define TSF_AUTO_CLOSE       fb_2 // Schliessen der Schnittstelle, nach Abarbeitung aller Aufgaben.

        // Klasse fuer die Verwaltung von Slots
        class Protocol_dockingstation_v1
        {
        public:
            enum
            {
                dv_write_trials = 3,
                dv_read_trials = 3
            };
        public:
            Protocol_dockingstation_v1()
            {
                // Member initialisiseren.
                m_connection = nullptr;

                // Verbindungsthread
                m_thread_state = ts_end;
                m_thread_command = tc_none;

                // Anzahl der Suspendanforderungen halten
                m_suspend_counter = 0;

                // Flags fuer Steuerung der Idlezeit
                m_flags = 0;
            }
            virtual ~Protocol_dockingstation_v1();

        public:
            // Setzt die Basisklasse fuer die Kommunikationsschicht
            void set_connection( Connection_base& connection )
            {
                m_connection = &connection;
            }
            Connection_base& get_connection() 
            { 
                if ( m_connection == nullptr )
                {
                    throw; 
                }

                return (Connection_base&)*m_connection; 
            }

            inline int is_slot( unsigned char device_address )
            {
                return (read_slot_firmware_version( device_address ) > 0);
            }

            unsigned char get_buffer_size( unsigned char device_address )
            {
                // Bereich pruefen
                if ( device_address < 1 && device_address > 255 )
                {
                    return 0;
                }

                // Je nach enthaltener Version
                switch ( m_timeboy_slot[device_address].get_version() ) 
                {
                case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
                    break;

                case 0: // Ermittlung noch nicht durchgefuehrt
                    return MB_SLOT_BUFFER_SIZE_SMALL;

                default:
                    if ( m_timeboy_slot[device_address].get_version() < 0x0131 )
                    {
                        return MB_SLOT_BUFFER_SIZE_SMALL;
                    }
                    else if ( m_timeboy_slot[device_address].get_version() >= 0x0131 )
                    {
                        return MB_SLOT_BUFFER_SIZE_LARGE;
                    }
                }

                return 0;
            }
            unsigned char get_data_buffer_size( unsigned char device_address )
            {
                // Bereich pruefen
                if ( device_address < 1 && device_address > 255 )
                {
                    return 0;
                }

                switch ( m_timeboy_slot[device_address].get_version() ) 
                {
                case 0xFFFF: // Pruefung bereits durchgefuehrt und war nicht erfolgreich.
                    break;

                case 0: // Ermittlung noch nicht durchgefuehrt
                    return MB_SLOT_BUFFER_DATA_SIZE_SMALL;

                default:
                    if ( m_timeboy_slot[device_address].get_version() < 0x0131 )
                    {
                        return MB_SLOT_BUFFER_DATA_SIZE_SMALL;
                    }
                    else if ( m_timeboy_slot[device_address].get_version() >= 0x0131 )
                    {
                        return MB_SLOT_BUFFER_DATA_SIZE_LARGE;
                    }
                }

                return 0;
            }

            // Main - Funktion fuer Befehlsausfuehrung in Workerthread
        protected:
            void communication_thread();
            std::thread m_communication_thread;

        protected:
            // Aktueller interner Zustand des Threads halten
            volatile enum enum_thread_state 
            {
                ts_begin, // Bei Einstieg in Main-Routine
                ts_end,   // Bei Ausstieg aus Main-Routine
                ts_idle,  // Thread befindet sich im Leerlauf (laufende Pruefung auf anstehende Aufgaben)
                ts_execute_command, // Abarbeitung eines gesetzten Kommandos (threadCommand != none)
                ts_dummy
            } m_thread_state;
        public:
            enum_thread_state get_thread_state()
            {
                return m_thread_state;
            }
        protected:
            // Auszufuehrendes Kommando
            volatile enum enum_thread_command
            {
                tc_none,  // Aktuell kein abzuarbeitender Befehl
                tc_exit,  // Beendigung des Threads erfordert
                tc_write, // Schreiben von Bytes erfordert (Daten in _commandParam)
                tc_read   // Lesen von Bytes erfordert (Daten in _commandParam)
            } m_thread_command;
        public:
            enum_thread_command get_thread_command()
            {
                return m_thread_command;
            }
            int set_thread_command( enum_thread_command tc )
            {
                if ( m_thread_state == ts_idle )
                {
                    m_thread_command = tc;
                    return 1;
                }

                return 0;
            }
            volatile struct  
            {
                int res;
                unsigned char device_address_receiver;
                unsigned int base_address;
                int offset;
                unsigned char* buffer;
                int count;
                int trials;
            } m_command_parameter;
            volatile int m_suspend_counter;
        protected:
            int start_communication_thread();
            void stop_communication_thread();

            // Gibt die FWVersion des Dockingslots an
            unsigned short read_slot_firmware_version( unsigned char device_address );
        public:
            void reset_slot( unsigned char device_address );
            void reset_slots();

        public: // Routinen fuer die Kommunikation mit Docking
            int write_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int write_trials = dv_write_trials );
            int read_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int read_trials = dv_read_trials );
        protected:
            int thread_write_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int write_trials = dv_write_trials );
            int thread_read_bytes( unsigned char device_address, unsigned int base_address, int offset, unsigned char* buffer, int count, int read_trials = dv_read_trials );

        public:
            int read_timeboy_model( unsigned char device_address, int& model );
            int set_timeboy_slot_state( unsigned char device_address, int state, int command_result = IV_ERROR_SUCCESS );

        protected:
            int call_docking( unsigned char* send_buffer, unsigned char send_count, unsigned char* receive_buffer, unsigned char* receive_count );
            int set_docking_led( unsigned char device_address, int state );

        protected: // Routinen fuer die Flagverwaltung
            volatile unsigned int m_flags;
        public:
            enum enum_flag_bit
            {
                fb_1  = 0x00000001, fb_2  = 0x00000002, fb_3  = 0x00000004, fb_4  = 0x00000008, fb_5  = 0x00000010, fb_6  = 0x00000020, fb_7  = 0x00000040, fb_8  = 0x00000080,
                fb_9  = 0x00000100, fb_10 = 0x00000200, fb_11 = 0x00000400, fb_12 = 0x00000800, fb_13 = 0x00001000, fb_14 = 0x00002000, fb_15 = 0x00004000, fb_16 = 0x00008000,
                fb_17 = 0x00010000, fb_18 = 0x00020000, fb_19 = 0x00040000, fb_20 = 0x00080000, fb_21 = 0x00100000, fb_22 = 0x00200000, fb_23 = 0x00400000, fb_24 = 0x00800000,
                fb_25 = 0x01000000, fb_26 = 0x02000000, fb_27 = 0x04000000, fb_28 = 0x08000000, fb_29 = 0x10000000, fb_30 = 0x20000000, fb_31 = 0x40000000, fb_32 = 0x80000000
            };
            void add_flag( enum_flag_bit flag )
            {
                m_flags |= flag;
            }
            void del_flag( enum_flag_bit flag )
            {
                m_flags &= ~flag;
            }
            int is_flag( enum_flag_bit flag )
            {
                return ((m_flags & flag) != 0);
            }

        protected:
            int m_auto_led_timeout;
        public:
            void set_led_auto_control( int enable, int timeout )
            {
                m_auto_led_timeout = timeout;
                if ( enable )
                {
                    add_flag( TSF_AUTO_LED );
                }
                else
                {
                    del_flag( TSF_AUTO_LED );
                }
            }

        protected:
            Connection_base* m_connection;
            Timeboy_slot m_timeboy_slot[255];

            // Aktuell bearbeitete DeviceID halten
        protected:
            unsigned char m_device_address_receiver;
            inline void set_device_address_receiver( unsigned char device_address )
            {
                m_device_address_receiver = device_address;
            }
            inline unsigned char device_address_receiver()
            {
                return m_device_address_receiver;
            }

            // Helper zur Paketerstellung
        protected:
            int bin_to_uint_msb_first( unsigned char* buffer )
            {
                return int(buffer[0] << 8) | int(buffer[1]);
            }
            unsigned int bin_to_ulong_msb_first( unsigned char* buffer )
            {
                unsigned int v;

                v  = buffer[0];
                v <<= 8;

                v |= buffer[1];
                v <<= 8;

                v |= buffer[2];
                v <<= 8;

                v |= buffer[3];

                return v;
            }

            unsigned short get_crc16( unsigned char* data, unsigned int size, unsigned short crc = 0xFFFF );
        };

    } // namespace Communication
} // namespace Datafox

#endif
